def rs_get_public_key(private_key: str) -> str: ...
def rs_pedersen_hash(left: str, right: str) -> str: ...
def rs_sign(msg_hash: str, private_key: str, k: str) -> tuple[str, str]: ...
def rs_verify(msg_hash: str, sig_r: str, sig_s: str, public_key: str) -> bool: ...
